#include "pch.h"
#include "Interfaces.h"
#include "Renderer.h"
#include "ShapeGeometry.h"

using namespace DrawingLayer;
// Properties

int Renderer::TargetWidth::get()
{
	return m_targetWidth;
}
int Renderer::TargetHeight::get()
{
	return m_targetHeight;
}
ImageBrush^ Renderer::TargetBrush::get()
{
	return m_imageBrush;
}
bool Renderer::EnableMotionBlur::get()
{
	return m_motionBlurEnabled;
}
void Renderer::EnableMotionBlur::set(bool value)
{
	m_motionBlurEnabled = value;
}
float Renderer::MotionBlurAngle::get()
{
	return m_motionBlurAngle;
}
void Renderer::MotionBlurAngle::set(float value)
{
	m_motionBlurAngle = value;
}
float Renderer::MotionBlurDeviation::get()
{
	return m_motionBlurDeviation;
}
void Renderer::MotionBlurDeviation::set(float value)
{
	m_motionBlurDeviation = value;
}


// Constructor
Renderer::Renderer()
{
    D2D1_FACTORY_OPTIONS options;
	ZeroMemory(&options, sizeof(options));
	ThrowIfFailed(D2D1CreateFactory(D2D1_FACTORY_TYPE_SINGLE_THREADED, __uuidof(ID2D1Factory1), &options, &m_d2dFactory));
    m_drawingInProcess = false;
    m_targetWidth = 0;
    m_targetHeight = 0;
	m_motionBlurDeviation = 10.0f;
}

// Private methods
void Renderer::PrepareSurface(ID2D1Bitmap1** ppBitmap)
{
	ComPtr<IDXGISurface> surface;
	RECT updateRect = { 0, 0, m_targetWidth, m_targetHeight };
    POINT offset;
	ThrowIfFailed(m_imageSourceNative->BeginDraw(updateRect, &surface, &offset));
	ThrowIfFailed(m_d2dContext->CreateBitmapFromDxgiSurface(surface.Get(), NULL, ppBitmap));	
}
D2D1_COLOR_F Renderer::GetD2DColor(Color color)
{
	D2D1_COLOR_F d2dColor =
	{
		(float)color.R / 255.0f,
		(float)color.G / 255.0f,
		(float)color.B / 255.0f,
		(float)color.A / 255.0f,
	};
	return d2dColor;
}

// IRenderer
void Renderer::Reset(int width, int height)
{
    D3D_FEATURE_LEVEL featureLevels[] = 
	{
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0,
		D3D_FEATURE_LEVEL_10_1,
		D3D_FEATURE_LEVEL_10_0,
		D3D_FEATURE_LEVEL_9_3,
		D3D_FEATURE_LEVEL_9_2,
		D3D_FEATURE_LEVEL_9_1
	};
    
    ThrowIfFailed(D3D11CreateDevice(nullptr, D3D_DRIVER_TYPE_HARDWARE, 0, D3D11_CREATE_DEVICE_BGRA_SUPPORT, featureLevels, 
									ARRAYSIZE(featureLevels), D3D11_SDK_VERSION, &m_d3dDevice, NULL, &m_d3dContext));
	ThrowIfFailed(m_d3dDevice.As(&m_dxgiDevice));
    ThrowIfFailed(m_d2dFactory->CreateDevice(m_dxgiDevice.Get(), &m_d2dDevice));	
	ThrowIfFailed(m_d2dDevice->CreateDeviceContext(D2D1_DEVICE_CONTEXT_OPTIONS_NONE, &m_d2dContext));

    m_imageSource = ref new SurfaceImageSource(width, height, false);
	IInspectable* inspectable = (IInspectable*) reinterpret_cast<IInspectable*>(m_imageSource);
    ThrowIfFailed(inspectable->QueryInterface(__uuidof(ISurfaceImageSourceNative), (void**)&m_imageSourceNative));
	ThrowIfFailed(m_imageSourceNative->SetDevice(m_dxgiDevice.Get()));
    
    m_imageBrush = ref new ImageBrush();
	m_imageBrush->ImageSource = m_imageSource;
    
    m_targetWidth = width;
	m_targetHeight = height;

	ThrowIfFailed(m_d2dContext->CreateEffect(CLSID_D2D1DirectionalBlur, &m_blurEffect));
    
    m_d2dContext->CreateBitmap(D2D1::SizeU(m_targetWidth, m_targetHeight), nullptr, 0, 
							   D2D1::BitmapProperties1(D2D1_BITMAP_OPTIONS_TARGET,D2D1::PixelFormat(DXGI_FORMAT_B8G8R8A8_UNORM, D2D1_ALPHA_MODE_PREMULTIPLIED)),
                               &m_inputSource);

    m_blurEffect->SetInput(0, m_inputSource.Get());
}
void Renderer::BeginDraw()
{
    if (!m_drawingInProcess && (m_d2dContext.Get() != NULL)) 
    {
        m_d2dContext->BeginDraw();	
        m_d2dContext->SetTarget(m_inputSource.Get());
        m_drawingInProcess = true;
    }
}
void Renderer::EndDraw()
{
    if (m_drawingInProcess)
    {
        m_d2dContext->EndDraw();
        
        ComPtr<ID2D1Bitmap1> bitmap;	
        PrepareSurface(&bitmap);
        m_d2dContext->SetTarget(bitmap.Get());
        m_d2dContext->BeginDraw();
		m_d2dContext->Clear(D2D1::ColorF(0.0f, 0.0f, 0.0f, 0.0f));
			
		if (m_motionBlurEnabled) 
		{
			m_blurEffect->SetValue(D2D1_DIRECTIONALBLUR_PROP_STANDARD_DEVIATION, m_motionBlurDeviation);
			m_blurEffect->SetValue(D2D1_DIRECTIONALBLUR_PROP_ANGLE, m_motionBlurAngle);
			m_d2dContext->DrawImage(m_blurEffect.Get());
		}
		else
			m_d2dContext->DrawImage(m_inputSource.Get());
        m_d2dContext->EndDraw();
        
        m_d2dContext->SetTarget(NULL);
		m_imageSourceNative->EndDraw();	
        m_drawingInProcess = false;
    }
}
void Renderer::Clear(Color color)
{
    if (!m_drawingInProcess)
        return;
    m_d2dContext->Clear(GetD2DColor(color));
}
IShapeGeometry^ Renderer::CreateGeometry()
{
    ComPtr<ID2D1PathGeometry1> geometry;
    m_d2dFactory->CreatePathGeometry(&geometry);
    return ref new ShapeGeometry(geometry.Get());
}
void Renderer::DrawGeometry(IShapeGeometry^ geometry, Color color, float width)
{
    if (!m_drawingInProcess)
        return;
    ShapeGeometry^ geometryObject = dynamic_cast<ShapeGeometry^>(geometry);
    geometryObject->Draw(m_d2dContext.Get(), GetD2DColor(color), width);
}
void Renderer::FillGeometry(IShapeGeometry^ geometry, Color color)
{
    if (!m_drawingInProcess)
        return;
    ShapeGeometry^ geometryObject = dynamic_cast<ShapeGeometry^>(geometry);
    geometryObject->Fill(m_d2dContext.Get(), GetD2DColor(color));
}


